from .exc import PermissionsError


class PermissionsMeta(type):

    """Applies permissions to class-based view methods.

    Does so by looking for a ``permissions`` dict on the view class. If
    it exists, the ``@permissions.require`` decorator will be applied
    to the specified methods. Here's an example::

        permissions = {
            'get': 'can_view_thing',
            'post': 'can_create_thing',
            'put': 'can_edit_thing',
        }

    This is handy for cases where you want to require a permission on
    on a super class method but don't want to override it just to add
    require the permission. I.e., it lets you avoid this::

        @permissions.require('permission_name')
        def some_method(self, *args, **kwargs):
            super().some_method(*args, **kwargs)

    The most convenient way to make of use of this is via the
    ``metaclass`` property of an existing registry::

        permissions_registry = PermissionsRegistry()

        class MyView(View, metaclass=permissions_registry.metaclass):

            permissions = {
                'get': 'can_view_stuff',
                'post': 'can_create_stuff',
            }

    Another option is to set a ``permissions_registry`` attribute on the
    view class.

    The final option is to make a subclass of :class:`PermissionsMeta`
    with a registry attribute. `PermissionsRegistry.metaclass` creates
    a registry configured in this way.

    """

    def __new__(mcs, name, bases, attrs):
        cls = type.__new__(mcs, name, bases, attrs)
        if hasattr(cls, 'permissions'):
            if hasattr(cls, 'permissions_registry'):
                registry = cls.permissions_registry
            elif hasattr(mcs, 'registry'):
                registry = mcs.registry
            else:
                raise PermissionsError('No permissions registry found')
            for k, v in cls.permissions.items():
                method = getattr(cls, k)
                decorated_method = registry.require(v)(method)
                setattr(cls, k, decorated_method)
        return cls



from django.test import TestCase

from django.template import Context, Template

from permissions import PermissionsRegistry


class Model:

    pass


def can_do(user):
    return user is not None


def can_do_with_model(user, instance):
    return None not in (user, instance)


class TestTemplateTags(TestCase):

    def setUp(self):
        self.registry = PermissionsRegistry()
        self.registry.register(can_do)
        self.registry.register(can_do_with_model, model=Model)
        self.template = Template(
            '{% load permissions %}'
            '{% if user|can_do %}can_do{% endif %}'
            '{% if user|can_do_with_model:instance %}can_do_with_model{% endif %}'
        )

    def test_can_do(self):
        context = Context({'user': object(), 'instance': None})
        result = self.template.render(context)
        self.assertIn('can_do', result)

    def test_cannot_do(self):
        context = Context({'user': None, 'instance': None})
        result = self.template.render(context)
        self.assertNotIn('can_do', result)

    def test_can_do_with_model(self):
        context = Context({'user': object(), 'instance': object()})
        result = self.template.render(context)
        self.assertIn('can_do_with_model', result)

    def test_cannot_do_with_model(self):
        context = Context({'user': None, 'instance': object()})
        result = self.template.render(context)
        self.assertNotIn('can_do_with_model', result)



import sys

from .exc import NoSuchPermissionError
from .registry import PermissionsRegistry


__all__ = ['permission']


registry = None


def permission(perm_func=None, **kwargs):
    """Add a permission to the global registry.

    Optional keyword args:

        - ``model`` A model class
        - ``allow_anonymous`` Normally, anonymous users will be forced
          to log in; setting this disables that
        - ``name`` A name for the permission; this defaults to
          ``perm_func.__name__``
        - ``replace`` If a permission has already been registered with
          a given name, an error will be raised unless this is set

    This decorator registers a permission function for use in templates
    and creates a decorator for it, which can be used on a Django view.
    This function can be used in two ways::

        # some/app/perms.py
        from permissions import permission

        @permission
        def can_do_something(user):
            ...

    Or::

        @permission(model=Car)
        def can_do_something_to_a_car(user, car):
            ...

    In the latter case, when can_do_something_to_a_car is used as a
    Django view decorator, the second argument to the view function is
    assumed to be the PK of a Car model object. That object is looked
    up, and passed to the permission function.

    This function automagically adds a ``decorators`` global to the
    module where the permission function being registered is defined.
    The ``decorators`` global is a simple object container for all the
    permissions defined in the module. These decorators can be used on
    views like so::

        # some/app/views.py
        from .perms import decorators

        @decorators.can_do_stuff
        def some_view(request):
            ...

    .. note:: Permission functions *cannot* be used directly to decorate
              views. You must import and access the view decorators via
              the ``decorators`` global as shown above.

    """
    global registry
    if registry is None:
        registry = PermissionsRegistry()
    if perm_func is None:
        return lambda f: permission(f, **kwargs)
    entry = registry.register(perm_func, _return_entry=True, **kwargs)
    perm_module = sys.modules[perm_func.__module__]
    if not hasattr(perm_module, 'decorators'):
        setattr(perm_module, 'decorators', DecoratorContainer())
    setattr(perm_module.decorators, entry.name, entry.view_decorator)
    return perm_func


class DecoratorContainer:

    """This is just a simple container for decorators.

    An instance of this class is tacked onto the module a permision
    function comes from. Its attributes are set dynamically to return
    the permission function as a decorator.

    """

    def __getattr__(self, name):
        # If this gets invoked, it means the permission identified by
        # ``name`` wasn't found in the "usual" places--that is, it's not
        # an attribute of the container instance, its class, super
        # classes, etc.
        raise NoSuchPermissionError(name)



from django.contrib.auth.models import AnonymousUser, User
from django.core.exceptions import PermissionDenied

from permissions import decorator, permission
from permissions.exc import DuplicatePermissionError, NoSuchPermissionError

from .base import Model, TestCase


class TestDecorator(TestCase):

    def setUp(self):
        super(TestDecorator, self).setUp()
        # Reset global registry on every run
        decorator.registry = self.registry

    def tearDown(self):
        globals().pop('decorators', None)

    def test_decoration(self):

        def can_do_stuff(user):
            return user.can_do_stuff

        original_can_do_stuff = can_do_stuff

        self.assertNotIn('decorators', globals())

        perm_func = permission(can_do_stuff, allow_anonymous=True)

        self.assertIs(original_can_do_stuff, perm_func)
        self.assertTrue(hasattr(decorator.registry, 'can_do_stuff'))
        self.assertIn('decorators', globals())

        @decorators.can_do_stuff
        def view(request):
            pass

        request = self.request_factory.get('/stuff')
        request.user = User()
        request.user.can_do_stuff = True
        view(request)
        request.user.can_do_stuff = False
        self.assertRaises(PermissionDenied, view, request)

    def test_decoration_with_model(self):

        @decorator.permission(model=Model, allow_anonymous=True)
        def can_do_things(user, instance):
            self.assertIsInstance(instance, Model)
            return user.can_do_things

        self.assertTrue(hasattr(decorator.registry, 'can_do_things'))

        @decorators.can_do_things
        def view(request, model_id):
            pass

        request = self.request_factory.get('/things/1')
        request.user = AnonymousUser()
        request.user.can_do_things = True
        view(request, model_id='1')

    def test_register_duplicate(self):
        decorator.permission(lambda u: None, name='perm')
        self.assertRaises(
            DuplicatePermissionError, decorator.permission, lambda u: None, name='perm')

    def test_unknown_permission(self):
        decorator.permission(lambda u: None, name='perm')
        with self.assertRaises(NoSuchPermissionError):
            decorators.huh(lambda r: None)



class PermissionsError(Exception):

    pass


class DuplicatePermissionError(PermissionsError):

    def __init__(self, name):
        super(DuplicatePermissionError, self).__init__('Permission exists: {0}'.format(name))


class NoSuchPermissionError(PermissionsError):

    def __init__(self, name):
        super(NoSuchPermissionError, self).__init__('Permission not registered: {0}'.format(name))



import inspect
import logging
from collections import namedtuple
from functools import wraps

import django.conf
from django.core.exceptions import PermissionDenied
from django.http import HttpRequest
from django.shortcuts import get_object_or_404
try:
    from django.utils.module_loading import import_string
except ImportError:
    from django.utils.module_loading import import_by_path as import_string

from .exc import DuplicatePermissionError, NoSuchPermissionError, PermissionsError
from .meta import PermissionsMeta
from .templatetags.permissions import register


log = logging.getLogger(__name__)


Entry = namedtuple('Entry', (
    'name', 'perm_func', 'view_decorator', 'model', 'allow_staff', 'allow_superuser',
    'allow_anonymous', 'unauthenticated_handler', 'request_types', 'views'
))


NO_VALUE = object()


DEFAULT_SETTINGS = {
    'allow_staff': False,
    'allow_superuser': False,
    'allow_anonymous': False,
    'unauthenticated_handler': None,
    'request_types': (),  # django.http.HttpRequest is always included
}


def _default(v, default):
    if v is None:
        return default
    return v


class PermissionsRegistry:

    """A registry of permissions.

    Args:

        - allow_staff: Allow staff to access all views by default. If
          this is set and the user is a staff member, the permission
          logic will not be invoked. [False]

        - allow_superuser: Allow superusers to access all views by
          default. If this is set and the user is a superuser, the
          permission logic will not be invoked. [False]

        - allow_anonymous: Allow anonymous users. Note: this is
          different from the two options above in that it doesn't
          grant permission by default but instead just gives anonymous
          users a chance to access a view--the permission logic is still
          invoked. [False]

        - unauthenticated_handler: A function that handles unpermitted
          requests by anonymous users. It's called when a view doesn't
          allow anonymous users and also when anonymous users are
          allowed but the permission check fails. It takes the current
          request as its only arg; it should return a response object.
          [Default behavior is to redirect to the login page]

        - request_types: A list of the types of request objects used by
          your project. In a typical Django project, this won't need to
          be set, but if you're using, e.g., Django REST Framework,
          you'll need to add its request class to this list.
          [django.http.HttpRequest]

          Note: You never need to add Django's request class to the
          ``request_types`` list; it will be added automatically if it's
          not present.

        If an option's value isn't passed to the constructor, it will
        be pulled from your project's settings or fall back to the
        defaults noted above in brackets.

        All options can be overridden on a per-permission basis by
        passing the corresponding argument to :meth:`register`.

    Create a registry somewhere in your project::

        # my/project/perms.py
        from permissions import PermissionsRegistry

        permissions = PermissionsRegistry()

    Then register permissions for an app like so::

        # my/project/app/perms.py
        from my.project.perms import permissions

        @permissions.register
        def can_do_stuff(user):
            ...

        @permissions.register(model=MyModel)
        def can_do_things(user, instance):
            ...

    Then require permissions on views like this::

        # my/project/app/views.py
        from my.project.perms import permissions

        @permissions.require('can_do_stuff')
        def my_view(request):
            ...

    TODO: Write more documentation.

    """

    def __init__(self, allow_staff=None, allow_superuser=None, allow_anonymous=None,
                 unauthenticated_handler=None, request_types=None):
        self._registry = dict()

        settings = DEFAULT_SETTINGS.copy()
        if hasattr(django.conf.settings, 'PERMISSIONS'):
            settings.update(django.conf.settings.PERMISSIONS)

        self._allow_staff = _default(allow_staff, settings['allow_staff'])
        self._allow_superuser = _default(allow_superuser, settings['allow_superuser'])
        self._allow_anonymous = _default(allow_anonymous, settings['allow_anonymous'])

        unauthenticated_handler = _default(
            unauthenticated_handler, settings['unauthenticated_handler'])

        if unauthenticated_handler is None:
            # Set up the default handler for unauthenticated requests.

            # Putting this import here is a hack-around for testing.
            # Merely importing login_required causes
            # django.conf.settings to be accessed in some other module,
            # which causes ImproperlyConfigured to be raised during the
            # import phase of test discovery.
            from django.contrib.auth.decorators import login_required

            # A fake view that, when called with the current request,
            # triggers Django's redirect-to-login functionality.
            force_login_view = login_required(lambda _: None)
            unauthenticated_handler = lambda r: force_login_view(r)
        else:
            if isinstance(unauthenticated_handler, str):
                unauthenticated_handler = import_string(unauthenticated_handler)
        self._unauthenticated_handler = unauthenticated_handler

        request_types = _default(request_types, settings['request_types'])
        request_types = tuple(import_string(t) for t in request_types if isinstance(t, str))
        if HttpRequest not in request_types:
            request_types = (HttpRequest,) + (request_types or ())
        self._request_types = request_types

    @property
    def metaclass(self):
        """Get a metaclass configured to use this registry."""
        if '_metaclass' not in self.__dict__:
            self._metaclass = type('PermissionsMeta', (PermissionsMeta,), {'registry': self})
        return self._metaclass

    def register(self, perm_func=None, model=None, allow_staff=None, allow_superuser=None,
                 allow_anonymous=None, unauthenticated_handler=None, request_types=None, name=None,
                 replace=False, _return_entry=False):
        """Register permission function & return the original function.

        This is typically used as a decorator::

            permissions = PermissionsRegistry()
            @permissions.register
            def can_do_something(user):
                ...

        For internal use only: you can pass ``_return_entry=True`` to
        have the registry :class:`.Entry` returned instead of
        ``perm_func``.

        """
        allow_staff = _default(allow_staff, self._allow_staff)
        allow_superuser = _default(allow_superuser, self._allow_superuser)
        allow_anonymous = _default(allow_anonymous, self._allow_anonymous)
        unauthenticated_handler = _default(unauthenticated_handler, self._unauthenticated_handler)
        request_types = _default(request_types, self._request_types)

        if perm_func is None:
            return (
                lambda perm_func_:
                    self.register(
                        perm_func_, model, allow_staff, allow_superuser, allow_anonymous,
                        unauthenticated_handler, request_types, name, replace, _return_entry)
            )

        name = _default(name, perm_func.__name__)
        if name == 'register':
            raise PermissionsError('register cannot be used as a permission name')
        elif name in self._registry and not replace:
            raise DuplicatePermissionError(name)

        view_decorator = self._make_view_decorator(
            name, perm_func, model, allow_staff, allow_superuser, allow_anonymous,
            unauthenticated_handler, request_types)
        entry = Entry(
            name, perm_func, view_decorator, model, allow_staff, allow_superuser, allow_anonymous,
            unauthenticated_handler, request_types, set())
        self._registry[name] = entry

        @wraps(perm_func)
        def filter_func(user, instance=NO_VALUE):
            test = lambda: perm_func(user) if instance is NO_VALUE else perm_func(user, instance)
            return (
                allow_staff and user.is_staff or
                allow_superuser and user.is_superuser or
                test()
            )

        register.filter(name, filter_func)

        log.debug('Registered permission: {0}'.format(name))
        return entry if _return_entry else perm_func

    __call__ = register

    def require(self, perm_name, **kwargs):
        """Use as a decorator on a view to require a permission.

        Optional args:

            - ``field`` The name of the model field to use for lookup
              (this is only relevant when requiring a permission that
              was registered with ``model=SomeModelClass``)

        Examples::

            @registry.require('can_do_stuff')
            def view(request):
                ...

            @registry.require('can_do_stuff_with_model', field='alt_id')
            def view_model(request, model_id):
                ...

        """
        view_decorator = self._get_entry(perm_name).view_decorator
        return view_decorator(**kwargs) if kwargs else view_decorator

    def __getattr__(self, name):
        return self.require(name)

    def _get_entry(self, perm_name):
        """Get registry entry for permission."""
        try:
            return self._registry[perm_name]
        except KeyError:
            raise NoSuchPermissionError(perm_name)

    def _get_view_name(self, view):
        """Get fully-qualified name for ``view``."""
        if hasattr(view, '__qualname__'):
            return view.__qualname__
        return '{0.__module__}.{0.__name__}'.format(view)

    def _make_view_decorator(self, perm_name, perm_func, model, allow_staff, allow_superuser,
                             allow_anonymous, unauthenticated_handler, request_types):

        # Putting this import here is a hack-around for testing. Merely
        # importing login_required causes django.conf.settings to be
        # accessed in some other module, which causes
        # ImproperlyConfigured to be raised during the import phase of
        # test discovery.
        from django.contrib.auth.decorators import login_required

        def view_decorator(view=None, field='pk'):

            if view is None:
                return lambda view_: view_decorator(view_, field)
            elif not callable(view):
                raise PermissionsError('Bad call to permissions decorator')

            entry = self._get_entry(perm_name)
            entry.views.add(self._get_view_name(view))

            # When a permission is applied to a class, which is presumed
            # to be a class-based view, instead apply the permission to
            # the class's dispatch() method. This will effectively
            # require the permission for all of the class's view methods:
            # get(), post(), etc. The class is returned as is.
            #
            # @permissions.require('can_do_stuff')
            # class MyView(View):
            #
            #     def get(request):
            #         ...
            #
            # In this example, the call to require() returns this
            # instance of view_decorator. When view_decorator is
            # called (via @), MyView is passed in. When the lines
            # below are reached, we decorate MyView.dispatch() and
            # then return MyView.
            if isinstance(view, type):
                view.dispatch = view_decorator(view.dispatch, field)
                return view

            # This contains the names of all of the view's args
            # (positional and keyword). This is used to find the field
            # value for permissions that operate on a model.
            view_args_spec = inspect.getargspec(view)
            view_arg_names = view_args_spec.args
            perm_func_arg_spec = inspect.getargspec(perm_func)
            perm_func_arg_names = perm_func_arg_spec.args

            @wraps(view)
            def wrapper(*args, **kwargs):
                # The following allows permissions decorators to work on
                # view functions and class-based view methods. Either
                # the first or the second arg must be the request. In
                # the latter case, the first arg will be an instance of
                # a class-based view).
                if isinstance(args[0], request_types):
                    request_index = 0
                elif isinstance(args[1], request_types):
                    request_index = 1
                else:
                    raise PermissionsError('Could not find request in args passed to view')

                request = args[request_index]
                user = request.user

                if not allow_anonymous and user.is_anonymous():
                    return unauthenticated_handler(request)

                def test():
                    # All this stuff is in this closure because it won't
                    # be needed if the permission check is bypassed. In
                    # particular, we want to avoid fetching the model
                    # instance if possible.
                    perm_func_args = [user]
                    perm_func_kwargs = {}

                    args_index = request_index + 1
                    remaining_args = args[args_index:]  # Args after request
                    remaining_arg_names = view_arg_names[args_index:]

                    view_args = kwargs.copy()
                    view_args['request'] = request
                    view_args.update(zip(remaining_arg_names, remaining_args))

                    if model is not None:
                        if remaining_args:
                            # Assume the 1st positional arg after the
                            # request passed to the view contains the
                            # field value...
                            field_val = remaining_args[0]
                        else:
                            # ...unless there are no positional args
                            # after the request; in that case, use the
                            # value of the first keyword arg.
                            field_val = kwargs[remaining_arg_names[0]]
                        instance = self._get_model_instance(model, **{field: field_val})
                        perm_func_args.append(instance)

                    # Starting after the perm func's required args
                    # (either user or user & instance), map view args
                    # to perm func args.
                    for n in perm_func_arg_names[len(perm_func_args):]:
                        if n in view_args:
                            perm_func_kwargs[n] = view_args[n]

                    return perm_func(*perm_func_args, **perm_func_kwargs)

                has_permission = (
                    allow_staff and user.is_staff or
                    allow_superuser and user.is_superuser or
                    test()
                )

                if has_permission:
                    return view(*args, **kwargs)
                elif user.is_anonymous():
                    return unauthenticated_handler(request)
                else:
                    # Tack on the permission name to the request for
                    # better error handling since Django doesn't
                    # give you access to the PermissionDenied
                    # exception object.
                    request.permission_name = perm_name
                    raise PermissionDenied(
                        'The "{0}" permission is required to access this resource'
                        .format(perm_name))

            return wrapper
        return view_decorator

    def entry_for_view(self, view, perm_name):
        """Get registry entry for permission if ``view`` requires it.

        In other words, if ``view`` requires the permission specified by
        ``perm_name``, return the :class:`Entry` associated with the
        permission. If ``view`` doesn't require the permission, return
        ``None`` instead.

        """
        view_name = self._get_view_name(view)
        entry = self._get_entry(perm_name)
        if view_name in entry.views:
            return entry
        return None

    def _get_model_instance(self, model, **kwargs):  # pragma: no cover
        return get_object_or_404(model, **kwargs)



from django.test import TestCase as BaseTestCase

from django.test import RequestFactory

from permissions import PermissionsRegistry as BasePermissionsRegistry


class PermissionsRegistry(BasePermissionsRegistry):

    def _get_model_instance(self, model, **kwargs):
        return model(**kwargs)


class Model:

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)


class TestCase(BaseTestCase):

    def setUp(self):
        self.registry = PermissionsRegistry()
        self.request_factory = RequestFactory()




import sys

from setuptools import setup, find_packages


django_version = '<1.7' if sys.version_info < (2, 7) else ''



setup(
    name='django-perms',
    version='1.1.0',
    url='https://github.com/PSU-OIT-ARC/django-perms',
    author='Matt Johnson',
    author_email='mdj2@pdx.edu',
    maintainer='Wyatt Baldwin',
    maintainer_email='wyatt.baldwin@pdx.edu',
    description='Syntactic sugar for handling permission functions in views, templates and in code',
    packages=find_packages(),
    zip_safe=False,
    extras_require={
        'dev': [
            'coverage',
            'django{version}'.format(version=django_version),
            'django-nose',
            'nose',
            'six',
        ],
    },
    classifiers=[
        'Framework :: Django',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ],
)



from django.conf.urls import patterns, url


urlpatterns = patterns(
    '',
    url(r'^things$', lambda r: None),
    url(r'^things/(\d+)$', lambda r: None),
    url(r'^stuff$', lambda r: None),
    url(r'^stuff/(\d+)$', lambda r: None),
)



from django.contrib.auth.models import User
from django.core.exceptions import PermissionDenied

import six

from ..exc import PermissionsError
from ..meta import PermissionsMeta

from .base import Model, TestCase


class TestMeta(TestCase):

    def setUp(self):
        super(TestMeta, self).setUp()

        @self.registry.register(model=Model)
        def can_view(user, instance):
            return user.can_view

    def _do_view_tests(self, View):
        view = View()
        request = self.request_factory.get('/things/1')
        request.user = User()
        request.user.can_view = False
        self.assertRaises(PermissionDenied, view.get, request, 1)
        request.user.can_view = True
        view.get(request, 1)

    def test_registry_metaclass(self):

        @six.add_metaclass(self.registry.metaclass)
        class View(object):

            permissions = {
                'get': 'can_view',
            }

            def get(self, request, model_id):
                pass

        self._do_view_tests(View)

    def test_specify_permissions_registry_on_view_class(self):

        @six.add_metaclass(PermissionsMeta)
        class View(object):

            permissions_registry = self.registry

            permissions = {
                'get': 'can_view',
            }

            def get(self, request, model_id):
                pass

        self._do_view_tests(View)

    def test_no_registry(self):

        with self.assertRaises(PermissionsError):
            @six.add_metaclass(PermissionsMeta)
            class View(object):

                permissions = {
                    'get': 'can_view',
                }

                def get(self, request, model_id):
                    pass



from django.contrib.auth.models import AnonymousUser, User
from django.core.exceptions import PermissionDenied
from django.views.generic import View

from permissions import PermissionsRegistry
from permissions.exc import NoSuchPermissionError, PermissionsError

from .base import Model, TestCase


class TestRegistry(TestCase):

    def test_register(self):

        @self.registry.register
        def can_do_things(user):
            pass

        self.assertTrue(hasattr(self.registry, 'can_do_things'))

        @self.registry.can_do_things
        def view(request):
            pass

    def test_register_with_args(self):

        @self.registry.register(model=Model, allow_anonymous=True)
        def can_do_things(user, instance):
            self.assertIsInstance(instance, Model)
            self.assertEqual(instance.model_id, 1)
            return user.can_do_things

        self.assertTrue(hasattr(self.registry, 'can_do_things'))

        @self.registry.require('can_do_things', field='model_id')
        def view(request, model_id):
            pass

        request = self.request_factory.get('/things/1')
        request.user = User()
        request.user.can_do_things = True
        view(request, 1)

    def test_cannot_use_register_as_perm_name(self):
        self.assertRaises(
            PermissionsError, self.registry.register, lambda u: None, name='register')

    def test_get_unknown_permission(self):
        with self.assertRaises(NoSuchPermissionError):
            self.registry.pants
        with self.assertRaises(NoSuchPermissionError):
            self.registry.require('pants')

    def test_bad_decoration(self):
        self.registry.register(lambda u: None, name='perm')
        self.assertRaises(PermissionsError, self.registry.perm, object())

    def test_apply_to_class_based_view(self):

        @self.registry.register(allow_anonymous=True)
        def can_do_things(user):
            return user.can_do_things

        @self.registry.require('can_do_things')
        class View(object):

            def dispatch(self, req):
                return getattr(self, req.method.lower())(req)

            def get(self, req):
                pass

        self.assertEqual(View.dispatch.__name__, 'dispatch')

        request = self.request_factory.get('/things')
        request.user = User()

        request.user.can_do_things = True
        view = View()
        view.dispatch(request)

        request.user.can_do_things = False
        self.assertRaises(PermissionDenied, view.dispatch, request)

    def test_apply_to_class_based_view_with_model(self):

        @self.registry.register(model=Model, allow_anonymous=True)
        def can_do_stuff(user, instance):
            return user.can_do_stuff and instance is not None

        @self.registry.require('can_do_stuff')
        class View(object):

            def dispatch(self, req, model_id, *args, **kwargs):
                return getattr(self, req.method.lower())(req, model_id, *args, **kwargs)

            def get(self, req, model_id):
                return model_id

        request = self.request_factory.get('/stuff/1')
        request.user = User()

        request.user.can_do_stuff = True
        view = View()
        view.dispatch(request, 1)

        request.user.can_do_stuff = False
        self.assertRaises(PermissionDenied, view.dispatch, request, model_id=1)

    def test_view_args_are_passed_through_to_perm_func(self):

        @self.registry.register
        def perm(user, model_id, request=None, not_a_view_arg='XXX'):
            self.assertEqual(model_id, 1)
            self.assertIs(request, request_passed_to_view)
            self.assertEqual(not_a_view_arg, 'XXX')
            return True

        @self.registry.perm
        def view(request, model_id, view_arg_that_is_not_passed_through):
            pass

        request_passed_to_view = self.request_factory.get('/things/1')
        request_passed_to_view.user = User()
        view(request_passed_to_view, 1, 2)

    def test_perm_func_is_not_called_when_user_is_staff_and_allow_staff_is_set(self):
        registry = PermissionsRegistry(allow_staff=True)

        @registry.register
        def perm(user):
            raise PermissionsError('Should not be raised')

        @registry.perm
        def view(request):
            pass

        request = self.request_factory.get('/things/1')
        request.user = User(is_staff=True)
        view(request)

    def test_anon_is_required_to_login(self):
        @self.registry.register
        def perm(user):
            return False

        @self.registry.require('perm')
        def view(request):
            pass

        request = self.request_factory.get('/things/1')
        request.user = AnonymousUser()
        response = view(request, 1)
        self.assertEqual(response.status_code, 302)

    def test_anon_is_required_to_login_when_perm_check_fails(self):
        @self.registry.register(allow_anonymous=True)
        def perm(user):
            return False

        @self.registry.require('perm')
        def view(request):
            pass

        request = self.request_factory.get('/things/1')
        request.user = AnonymousUser()
        response = view(request, 1)
        self.assertEqual(response.status_code, 302)

    def test_ensure_custom_unauthenticated_handler_is_called(self):
        def handler(request):
            handler.called = True
        handler.called = False

        registry = PermissionsRegistry(unauthenticated_handler=handler)

        @registry.register
        def perm(user):
            return False

        @registry.require('perm')
        def view(request):
            pass

        request = self.request_factory.get('/things/1')
        request.user = AnonymousUser()
        self.assertFalse(handler.called)
        view(request, 1)
        self.assertTrue(handler.called)

    def test_ensure_view_perms(self):

        @self.registry.register
        def perm(user):
            pass

        @self.registry.require('perm')
        def view(request):
            pass

        entry = self.registry.entry_for_view(view, 'perm')
        self.assertIsNotNone(entry)
        self.assertIs(entry.perm_func, perm)

        # try the same thing with a CBV
        @self.registry.require('perm')
        class AView(View):
            def get(self, request):
                pass

        entry = self.registry.entry_for_view(AView, 'perm')
        self.assertIsNotNone(entry)
        self.assertIs(entry.perm_func, perm)

        # same thing with the permission on a CBV method
        class AnotherView(View):
            @self.registry.require('perm')
            def get(self, request):
                pass

        entry = self.registry.entry_for_view(AnotherView.get, 'perm')
        self.assertIsNotNone(entry)
        self.assertIs(entry.perm_func, perm)
